{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Object Oriented Programming in Python\n",
    "#### by Jochen Hinz\n",
    "\n",
    "##### In this talk we will learn the basics of object oriented programming in Python in an intuitive way.\n",
    "##### No boring abstract stuff, just hands-on examples."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "##### To highlight the differences between object oriented and functional programming, let us start with a functional programming example that some of you may remember from the last BaNaNa talk about Python."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "def evaluate_polynomial( x, prefactors ):\n",
    "    \"\"\" evaluate polynomial with prefactors ``prefactors`` in ``x`` \"\"\"\n",
    "    return sum( a * x ** i for i, a in enumerate( prefactors ) )\n",
    "\n",
    "prefactors = [ 10, 2, 5, 6, 8 ]\n",
    "x = 2\n",
    "\n",
    "print( evaluate_polynomial( x, prefactors ) )"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "##### What if we want to allow for the evaluation of polynomials and their derivatives ?\n",
    "##### The functional programming solution:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "def evaluate_polynomial_with_derivative( x, prefactors, der=0 ):\n",
    "    \n",
    "    \"\"\" evaluate the ``der``-th derivative of a \n",
    "        polynomial with prefactors ``prefactors`` in ``x`` \"\"\"\n",
    "    \n",
    "    der = int(der)\n",
    "    assert der >= 0  # no negative derivatives\n",
    "    \n",
    "    if der > 0:\n",
    "        \n",
    "        \"\"\"``der`` is not 0 => take derivative: \n",
    "        d/dx(a0 + a1 x + a2 x^2 + ...) = a1 + 2*a2 x + ... \"\"\"\n",
    "        \n",
    "        derivative_prefactors = [ a * i for i, a in enumerate(prefactors) ][1:]\n",
    "        \n",
    "        return evaluate_polynomial_with_derivative( x, derivative_prefactors, der=der-1 )\n",
    "    \n",
    "    else:  ## der is 0: simply evaluate the polynomial\n",
    "        return sum( a * x ** i for i, a in enumerate( prefactors ) )\n",
    "\n",
    "prefactors = [ 10, 2, 5, 6, 8 ]\n",
    "x = 2\n",
    "\n",
    "print( '0-th derivative:', evaluate_polynomial_with_derivative( x, prefactors ) )\n",
    "print( '1st derivative:', evaluate_polynomial_with_derivative( x, prefactors, der=1 ) )\n",
    "print( '2nd derivative:', evaluate_polynomial_with_derivative( x, prefactors, der=2 ) )"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "##### A polynomial constitutes a particular >>class<< of functions. \n",
    "##### We make our first OOP-steps by writing a class that represents a polynomial.\n",
    "# 1. The Basics of OOP"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "\"\"\"\n",
    "    Don't be scared by this cell. We will break it down step-by-step.\n",
    "\"\"\"\n",
    "\n",
    "class Polynomial:\n",
    "    \n",
    "    ## The initializer\n",
    "    def __init__( self, prefactors ):\n",
    "        \n",
    "        print('This is conclusive proof that the __init__ class function has been called automatically')\n",
    "        print('However, what is this ``self`` argument ?')\n",
    "        print('Answer:', self)\n",
    "        print('What is the content of ``self`` ?')\n",
    "        print('Answer:', self.__dict__)  # __dict__ returns a dictionary with all attributes of self\n",
    "        print('Oh, so ``self`` is an empty Polynomial object !')\n",
    "        print('Let us set the prefactors by invoking self.prefactors = prefactors')\n",
    "        \n",
    "        self.prefactors = prefactors\n",
    "        \n",
    "        print('The content of ``self`` is now:', self.__dict__)  # print the content again\n",
    "        print(\"``self`` is now aware of its own prefactors !\")\n",
    "        print('At this point the __init__ class function terminates. \\n')\n",
    "        \n",
    "    def evaluate( self, x ):\n",
    "        print( \"We're in the ``evaluate`` class function now.\")\n",
    "        print( 'Inside the evaluate function the variable ``self`` refers to', self )\n",
    "        print( \"So Python has automatically passed ``mypolynomial`` as argument ``self`` to the evaluate function \\n\")\n",
    "        print( \"``self is aware of its own prefactors. Let's print self.prefactors\")\n",
    "        print( \"self.prefactors:\", self.prefactors, '\\n')\n",
    "        print( 'We can now return the polynomial evaluation by utilizing ``self.prefactors`` \\n')\n",
    "        return sum( a * x ** i for i, a in enumerate( self.prefactors ) ) # as before     \n",
    "\n",
    "\n",
    "prefactors = [ 10, 2, 5, 6, 8 ]\n",
    "x = 2\n",
    "\n",
    "print(\"Setting variable 'mypolynomial' to Polynomial(prefactors) with prefactors =\", prefactors, '\\n')\n",
    "mypolynomial = Polynomial( prefactors )\n",
    "\n",
    "print(\"What does the variable 'mypolynomial' refer to ?\")\n",
    "print('Answer:', mypolynomial)\n",
    "print(\"Oh, so Python has set what used to be ``self`` within ``__init__`` to the variable ``mypolynomial``. \\n\")\n",
    "print(\"The prefactors of 'mypolynomial' are:\", mypolynomial.prefactors, '\\n')\n",
    "\n",
    "print('Calling mypolinomial.evaluate(x) \\n')\n",
    "print( '0-th derivative in x=2:', mypolynomial.evaluate(x) )"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "##### This seems to work, but we can't take derivatives yet.\n",
    "##### Obviously, the derivative of a polynomial is a polynomial."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class PolynomialWithDerivative:\n",
    "    \n",
    "    def __init__( self, prefactors ):\n",
    "        self.prefactors = prefactors\n",
    "        \n",
    "    def evaluate( self, x ):\n",
    "        return sum( a * x ** i for i, a in enumerate( self.prefactors ) )  \n",
    "    \n",
    "    def derivative( self ):\n",
    "        \"\"\"Same trick as in the functional programming example\"\"\"\n",
    "        derivative_prefactors = [ a * i for i, a in enumerate(self.prefactors) ][1:]\n",
    "        \"\"\"We simply return a new Polynomial with the new prefactors\"\"\"\n",
    "        return PolynomialWithDerivative( derivative_prefactors )\n",
    "\n",
    "prefactors = [ 10, 2, 5, 6, 8 ]\n",
    "x = 2\n",
    "\n",
    "pol = PolynomialWithDerivative(prefactors)\n",
    "print( '0-th derivative:', pol.evaluate(x) )\n",
    "\n",
    "pol_der = pol.derivative()\n",
    "print( '1st derivative:', pol_der.evaluate(x) )\n",
    "\n",
    "pol_der_der = pol_der.derivative()\n",
    "print( '2nd derivative:', pol_der_der.evaluate(x) )"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "##### We can concatenate operations like this:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class PolynomialWithDerivative:\n",
    "    'No changes here'\n",
    "    \n",
    "    def __init__( self, prefactors ):\n",
    "        self.prefactors = prefactors\n",
    "        \n",
    "    def evaluate( self, x ):\n",
    "        return sum( a * x ** i for i, a in enumerate( self.prefactors ) )  \n",
    "    \n",
    "    def derivative( self ):\n",
    "        derivative_prefactors = [ a * i for i, a in enumerate(self.prefactors) ][1:]\n",
    "        return PolynomialWithDerivative( derivative_prefactors )\n",
    "\n",
    "prefactors = [ 10, 2, 5, 6, 8 ]\n",
    "x = 2\n",
    "\n",
    "pol = PolynomialWithDerivative(prefactors)\n",
    "\n",
    "print( '0-th derivative:', pol.evaluate(x) )\n",
    "print( '1st derivative:', pol.derivative().evaluate(x) )\n",
    "print( '2nd derivative:', pol.derivative().derivative().evaluate(x) )"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "##### MUCH MORE intuitive than the functional programming example !\n",
    "##### However, this .evaluate(x) is kinda stupid, let's do it differently:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class PolynomialWithDerivative:\n",
    "    \n",
    "    def __init__( self, prefactors ):\n",
    "        self.prefactors = prefactors\n",
    "        \n",
    "    def __call__( self, x ):  # replaces evaluate\n",
    "        return sum( a * x ** i for i, a in enumerate( self.prefactors ) )\n",
    "    \n",
    "    def derivative( self ):\n",
    "        derivative_prefactors = [ a * i for i, a in enumerate(self.prefactors) ][1:]\n",
    "        return PolynomialWithDerivative( derivative_prefactors )\n",
    "\n",
    "prefactors = [ 10, 2, 5, 6, 8 ]\n",
    "\n",
    "pol = PolynomialWithDerivative(prefactors)\n",
    "print( '0-th derivative:', pol(2) )\n",
    "print( '1st derivative:', pol.derivative()(2) )\n",
    "print( '2nd derivative:', pol.derivative().derivative()(2) )"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "##### Beautiful and intuitive !\n",
    "##### Another example: exponentials."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "\n",
    "class ExponentialWithDerivative:\n",
    "    \"Implements exp(a*x)\"\n",
    "    \n",
    "    def __init__( self, a ):\n",
    "        self.a = a\n",
    "    \n",
    "    def __call__( self, x ):\n",
    "        return np.exp( self.a * x )\n",
    "    \n",
    "    def derivative( self ):\n",
    "        return \"\"\"Dammit, this does not work because exp(a*x) -> a*exp(a*x) \"\"\"\n",
    "        \n",
    "exp = ExponentialWithDerivative( 4 )\n",
    "print( 'exp evaluated in 2:', exp(2) )\n",
    "print( 'derivative of exp:', exp.derivative() )"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "##### How are we gonna fix this ?\n",
    "##### Maybe like this:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "\n",
    "class ExponentialWithPrefactorAndDerivative:\n",
    "    \"Implements K*exp(a*x)\"\n",
    "    \n",
    "    def __init__( self, a, K=1 ):\n",
    "        self.a = a\n",
    "        self.K = K\n",
    "    \n",
    "    def __call__( self, x ):\n",
    "        return self.K * np.exp( self.a * x )\n",
    "    \n",
    "    def derivative( self ):\n",
    "        return ExponentialWithPrefactorAndDerivative( self.a, K=self.K*self.a )\n",
    "        \n",
    "exp = ExponentialWithPrefactorAndDerivative( 4 )\n",
    "print( 'exp evaluated in 2:', exp(2) )\n",
    "print( 'derivative of exp evaluated in 2:', exp.derivative()(2) )"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "##### It works but it is not general enough.\n",
    "##### In fact, Polynomial, Exponential and even the prefactor of the Exponential (K) are all functions.\n",
    "##### For more flexibility, this should be reflected in our code.\n",
    "##### This brings us to our next topic:\n",
    "# 2. Inheritance"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class FunctionWithDerivative:\n",
    "    'I am a template for other classes.'\n",
    "    \n",
    "    def __init__( self, function ):\n",
    "        self.function = function\n",
    "        \n",
    "    def __call__( self, x ):\n",
    "        return self.function(x)\n",
    "    \n",
    "    def derivative( self ):\n",
    "        raise AssertionError('Any particular class of function has to implement its derivative.')\n",
    "\n",
    "        \n",
    "class Constant( FunctionWithDerivative ):  ## the constant function inherits from FunctionWithDerivative\n",
    "    \n",
    "    def __init__( self, constant ):\n",
    "        \n",
    "        self.c = constant\n",
    "        function = lambda x: self.c ## will always return constant regardless of x\n",
    "        \n",
    "        \"\"\"\n",
    "            run the __init__ of FunctionWithDerivative (which is the super() of this class)\n",
    "            with the function we just made in order to set self.function\n",
    "        \"\"\"\n",
    "\n",
    "        super().__init__( function )\n",
    "        ## sets self.function = function (see FunctionWithDerivative)\n",
    "        \n",
    "    \"\"\"\n",
    "        We don't have to implement __call__ because this class will \n",
    "        inherit the behaviour of the base class (FunctionWithDerivative).\n",
    "    \"\"\"\n",
    "    \n",
    "    def derivative( self ):\n",
    "        \"\"\"\n",
    "            We do, however, overwrite the derivative which \n",
    "            was not specified in the base class.\n",
    "            d/dx(A) = 0\n",
    "        \"\"\"\n",
    "        return Constant(0)\n",
    "    \n",
    "    \n",
    "c = Constant( 5 )\n",
    "print( '0-th derivative in 2:', c(2) )\n",
    "print( '1st derivative in 2:', c.derivative()(2) )"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Polynomial( FunctionWithDerivative ):\n",
    "    \n",
    "    def __init__( self, prefactors ):\n",
    "        self.prefactors = prefactors\n",
    "        function = lambda x: \\\n",
    "            sum( a * x ** i for i, a in enumerate( self.prefactors ) )\n",
    "        super().__init__( function )\n",
    "    \n",
    "    def derivative( self ):\n",
    "        derivative_prefactors = [ a * i for i, a in enumerate(self.prefactors) ][1:]\n",
    "        return Polynomial( derivative_prefactors )\n",
    "    \n",
    "\n",
    "prefactors = [ 10, 2, 5, 6, 8 ]\n",
    "\n",
    "pol = Polynomial(prefactors)\n",
    "print( '0-th derivative:', pol(2) )\n",
    "print( '1st derivative:', pol.derivative()(2) )\n",
    "print( '2nd derivative:', pol.derivative().derivative()(2) )"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "##### Nice ! But how are we gonna solve the exp(a*x) -> a * exp(a*x) dilemma ?\n",
    "##### We need more ingredients:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Multiply( FunctionWithDerivative ):\n",
    "    'Function that returns the product of two other functions.'\n",
    "    \n",
    "    def __init__( self, func0, func1 ):\n",
    "        \n",
    "        # make sure both inputs are Functions with derivatives\n",
    "        assert \\\n",
    "            isinstance(func0, FunctionWithDerivative) \\\n",
    "            and isinstance(func1, FunctionWithDerivative)\n",
    "            \n",
    "        self.func0 = func0\n",
    "        self.func1 = func1\n",
    "        \n",
    "        function = lambda x: self.func0(x) * self.func1(x)\n",
    "        \n",
    "        super().__init__( function )\n",
    "        \n",
    "        \n",
    "class Exponential( FunctionWithDerivative ):\n",
    "    'exp(a*x)'\n",
    "    \n",
    "    def __init__( self, a ):\n",
    "        self.a = a\n",
    "        function = lambda x: np.exp( self.a * x )\n",
    "        super().__init__( function )\n",
    "        \n",
    "    def derivative( self ):\n",
    "        'd/dx(exp(a*x)) -> a*exp(a*x)'\n",
    "        return Multiply( Constant(self.a), Exponential(self.a) )\n",
    "\n",
    "    \n",
    "exp = Exponential( 4 )\n",
    "print( 'exp evaluated in 2:', exp(2) )\n",
    "print( 'derivative of exp evaluated in 2:', exp.derivative()(2) )"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "##### Awesome, but we have not implemented the derivative of a Multiply :-("
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "exp = Exponential( 4 )\n",
    "exp_der = exp.derivative()\n",
    "\n",
    "print( 'exp_der is a', exp_der )\n",
    "\n",
    "try:\n",
    "    print( 'Trying to compute the derivative of exp_der.' )\n",
    "    print(exp_der.derivative())\n",
    "except Exception as ex:\n",
    "    print( 'Failed with error: {}'.format(ex) )"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": true
   },
   "source": [
    "##### In order to implement the derivative of a Multiply, we need another ingredient"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Add( FunctionWithDerivative ):\n",
    "    'Implements the sum of two instances of FunctionWithDerivative'\n",
    "    \n",
    "    def __init__( self, func0, func1 ):\n",
    "        \n",
    "        assert \\\n",
    "            isinstance(func0, FunctionWithDerivative) \\\n",
    "            and isinstance(func1, FunctionWithDerivative)\n",
    "            \n",
    "        self.func0 = func0\n",
    "        self.func1 = func1\n",
    "        \n",
    "        function = lambda x: self.func0(x) + self.func1(x)\n",
    "        \n",
    "        super().__init__( function )\n",
    "        \n",
    "    def derivative( self ):\n",
    "        'd/dx( f + g ) = df/dx + dg/dx'\n",
    "        return Add( self.func0.derivative(), self.func1.derivative() )\n",
    "    \n",
    "\n",
    "class Multiply( FunctionWithDerivative ):\n",
    "    'Function that returns the product of two other functions.'\n",
    "    \n",
    "    def __init__( self, func0, func1 ):\n",
    "        \n",
    "        # make sure both inputs are Functions with derivatives\n",
    "        assert \\\n",
    "            isinstance(func0, FunctionWithDerivative) \\\n",
    "            and isinstance(func1, FunctionWithDerivative)\n",
    "            \n",
    "        self.func0 = func0\n",
    "        self.func1 = func1\n",
    "        \n",
    "        function = lambda x: func0(x) * func1(x)\n",
    "        \n",
    "        super().__init__( function )\n",
    "        \n",
    "    def derivative( self ):\n",
    "        \n",
    "        \"\"\"Product rule of differentiation:\n",
    "               d/dx( f * g ) = df/dx * g + f * dg/dx \"\"\"\n",
    "        \n",
    "        return Add(\n",
    "            Multiply(self.func0.derivative(), self.func1),\n",
    "            Multiply(self.func0, self.func1.derivative())\n",
    "        )"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "exp = Exponential( 4 )\n",
    "x = 2\n",
    "\n",
    "print( '0-th derivative of exp(4*x) in x=2:', exp(x) )\n",
    "print( '1st derivative of exp(4*x) in x=2:', exp.derivative()(x) )\n",
    "print( '2nd derivative of exp(4*x) in x=2:', exp.derivative().derivative()(x) )\n",
    "print( '3rd derivative of exp(4*x) in x=2:', exp.derivative().derivative().derivative()(x) )"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%reset"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Putting it all together:\n",
    "#### (You don't have to understand all the details)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import abc  ## abstract base class\n",
    "import numpy as np\n",
    "import numbers\n",
    "\n",
    "class FunctionWithDerivative(abc.ABC):\n",
    "    \n",
    "    def __init__( self, function ):\n",
    "        self.function = function\n",
    "        \n",
    "    def __call__( self, x ):\n",
    "        return self.function( x )\n",
    "    \n",
    "    @abc.abstractmethod\n",
    "    def _derivative(self):\n",
    "        'Implements the derivative.'\n",
    "        pass\n",
    "    \n",
    "    def derivative(self, n=1):\n",
    "        assert n >= 0\n",
    "        if n > 0:\n",
    "            return self._derivative().derivative(n - 1)\n",
    "        return self\n",
    "    \n",
    "    def __add__( self, other ):\n",
    "        if isinstance( other, numbers.Real ):\n",
    "            other = Constant(other)\n",
    "        if isinstance( other, FunctionWithDerivative ):\n",
    "            if isinstance( other, Zero ):\n",
    "                return self\n",
    "            return Add( self, other )\n",
    "        return NotImplemented\n",
    "    \n",
    "    def __radd__( self, other ):\n",
    "        return self.__add__(other)\n",
    "    \n",
    "    def __mul__( self, other ):\n",
    "        if isinstance( other, numbers.Real ):\n",
    "            other = Constant(other)\n",
    "        if isinstance( other, FunctionWithDerivative ):\n",
    "            if isinstance( other, Zero ):\n",
    "                return Zero()\n",
    "            return Multiply( self, other )\n",
    "        return NotImplemented\n",
    "    \n",
    "    def __rmul__( self, other ):\n",
    "        return self.__mul__(other)\n",
    "    \n",
    "    def plot( self, interval=None ):\n",
    "        if interval is None:\n",
    "            interval = [0, 1]\n",
    "        x = np.linspace(*interval, 100)\n",
    "        y = [ self(x_) for x_ in x ]\n",
    "        from matplotlib import pyplot as plt\n",
    "        %matplotlib inline\n",
    "        plt.figure()\n",
    "        plt.plot( x, y )\n",
    "        plt.show();\n",
    "        \n",
    "    \n",
    "class Add( FunctionWithDerivative ):\n",
    "    \n",
    "    def __init__( self, func0, func1 ):\n",
    "        \n",
    "        assert \\\n",
    "            isinstance(func0, FunctionWithDerivative) \\\n",
    "            and isinstance(func1, FunctionWithDerivative)\n",
    "            \n",
    "        self.func0 = func0\n",
    "        self.func1 = func1\n",
    "        \n",
    "        function = lambda x: func0(x) + func1(x)\n",
    "        \n",
    "        super().__init__( function )\n",
    "        \n",
    "    def _derivative( self ):\n",
    "        return self.func0.derivative() + self.func1.derivative()\n",
    "    \n",
    "\n",
    "class Multiply( FunctionWithDerivative ):\n",
    "    \n",
    "    def __init__( self, func0, func1 ):\n",
    "        \n",
    "        # make sure both inputs are Functions with derivatives\n",
    "        assert \\\n",
    "            isinstance(func0, FunctionWithDerivative) \\\n",
    "            and isinstance(func1, FunctionWithDerivative)\n",
    "            \n",
    "        self.func0 = func0\n",
    "        self.func1 = func1\n",
    "        \n",
    "        function = lambda x: func0(x) * func1(x)\n",
    "        \n",
    "        super().__init__( function )\n",
    "        \n",
    "    def _derivative( self ):\n",
    "        return self.func0.derivative() * self.func1 + \\\n",
    "                                     self.func0 * self.func1.derivative()\n",
    "\n",
    "    \n",
    "class Constant( FunctionWithDerivative ):\n",
    "    \n",
    "    def __new__( cls, constant ):\n",
    "        \"\"\"\n",
    "            if constant == 0 return the Zero function.\n",
    "            else do the usual.\n",
    "        \"\"\"\n",
    "        \n",
    "        if constant == 0:\n",
    "            return Zero()\n",
    "        \n",
    "        return FunctionWithDerivative.__new__(cls)    \n",
    "    \n",
    "    def __init__( self, constant ):\n",
    "        \n",
    "        self.c = constant\n",
    "        function = lambda x: constant\n",
    "        super().__init__( function )\n",
    "\n",
    "    def _derivative( self ):\n",
    "        return Constant(0)\n",
    "\n",
    "    \n",
    "class Zero( FunctionWithDerivative ):\n",
    "    \n",
    "    def __init__( self ):\n",
    "        super().__init__( lambda x: 0 )\n",
    "        \n",
    "    def _derivative( self ):\n",
    "        return Zero()\n",
    "        \n",
    "        \n",
    "class Exponential( FunctionWithDerivative ):\n",
    "    'exp(a*x)'\n",
    "    \n",
    "    def __new__( cls, a ):\n",
    "        if a == 0:\n",
    "            'a=0 -> exp(a*x) = 1'\n",
    "            return Constant(1)\n",
    "        return FunctionWithDerivative.__new__(cls)\n",
    "    \n",
    "    def __init__( self, a ):\n",
    "        self.a = a\n",
    "        function = lambda x: np.exp( self.a * x )\n",
    "        super().__init__( function )\n",
    "        \n",
    "    def _derivative( self ):\n",
    "        return self.a * Exponential(self.a)\n",
    "    \n",
    "    \n",
    "class Polynomial( FunctionWithDerivative ):\n",
    "    \n",
    "    def __new__( cls, prefactors ):\n",
    "        if len(prefactors) == 0:\n",
    "            return Zero()\n",
    "        if len(prefactors) == 1:\n",
    "            return Constant( prefactors[0] )\n",
    "        return FunctionWithDerivative.__new__(cls)\n",
    "    \n",
    "    def __init__( self, prefactors ):\n",
    "        self.prefactors = prefactors\n",
    "        function = lambda x: \\\n",
    "            sum( a * x ** i for i, a in enumerate( self.prefactors ) )\n",
    "        super().__init__( function )\n",
    "    \n",
    "    def _derivative( self ):\n",
    "        derivative_prefactors = [ a * i for i, a in enumerate(self.prefactors) ][1:]\n",
    "        return Polynomial( derivative_prefactors )\n",
    "\n",
    "    \n",
    "class Sin( FunctionWithDerivative ):\n",
    "    'sin(w*x)'\n",
    "    \n",
    "    def __new__( cls, w ):\n",
    "        if w == 0:\n",
    "            return Zero()\n",
    "        return FunctionWithDerivative.__new__(cls)\n",
    "    \n",
    "    def __init__( self, w ):\n",
    "        self.w = w\n",
    "        super().__init__( lambda x: np.sin( self.w * x) )\n",
    "        \n",
    "    def _derivative( self ):\n",
    "        return self.w * Cos( self.w )\n",
    "\n",
    "    \n",
    "class Cos( FunctionWithDerivative ):\n",
    "    'cos(w*x)'\n",
    "    \n",
    "    def __new__( cls, w ):\n",
    "        if w == 0:\n",
    "            return Constant(1)\n",
    "        return FunctionWithDerivative.__new__(cls)\n",
    "    \n",
    "    def __init__( self, w ):\n",
    "        self.w = w\n",
    "        super().__init__( lambda x: np.cos( self.w * x) )\n",
    "        \n",
    "    def _derivative( self ):\n",
    "        return -self.w * Sin( self.w )"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "exp = Exponential(-3)\n",
    "pol = Polynomial( [1, -1, 1, -1, 1] )\n",
    "sin = Sin(30)\n",
    "a = 4\n",
    "(a * exp * pol * sin + 10).plot()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Sanity check:\n",
    "$\\frac{\\mathrm{d}^2 y }{ \\mathrm{d} x ^ 2} + \\frac{\\mathrm{d} y }{ \\mathrm{d} x} + \\frac{5}{4}y = 0$\n",
    "### is solved by\n",
    "$y(x) = A \\mathrm{e}^{-\\frac{1}{2} x } \\sin{( x )} + B \\mathrm{e}^{-\\frac{1}{2} x } \\cos{( x )}$"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "A = 2\n",
    "B = -3\n",
    "y = A * Exponential(-1/2) * Sin(1) + B * Exponential(-1/2) * Cos(1)\n",
    "y.plot(interval=[0, 10])\n",
    "(y.derivative(2) + y.derivative(1) + 5/4 * y ).plot(interval=[0, 10])"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### How could we further improve above code ?"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<br><br><br><br><br><br><br>\n",
    "### Allow for function compositions by implementing the chain rule !\n",
    "##### (left as a homework exercise)\n",
    "<br><br><br><br><br>"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "#### Last topic (if time permits)\n",
    "# 3. Multiple Inheritance"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Is multiple inheritance a good idea ?\n",
    "### Short answer:\n",
    "#### No.\n",
    "### Long answer:\n",
    "#### \"If you think you need multiple inheritance, you're almost always wrong. If you know you need it, you're almost always right.\""
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### There exists one application of multiple inheritance that is considered 'good practice'\n",
    "## Mixins"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class FindRootMixin:\n",
    "    \n",
    "    def find_root( self, initial_guess=0 ):\n",
    "        from scipy import optimize\n",
    "        root = optimize.root( self, initial_guess )\n",
    "        if root.success:\n",
    "            print( 'Found a root at x={}'.format(root.x[0]) )\n",
    "            return root.x[0]\n",
    "        else:\n",
    "            print( 'No root found.' )"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import abc  ## abstract base class\n",
    "import numpy as np\n",
    "import numbers\n",
    "\n",
    "\"\"\"\n",
    "    Exactly the same as before with a little twist.\n",
    "\"\"\"\n",
    "\n",
    "class FunctionWithDerivative(abc.ABC):\n",
    "    \n",
    "    def __init__( self, function ):\n",
    "        self.function = function\n",
    "        \n",
    "    def __call__( self, x ):\n",
    "        return self.function( x )\n",
    "    \n",
    "    @abc.abstractmethod\n",
    "    def _derivative(self):\n",
    "        'Implements the derivative.'\n",
    "        pass\n",
    "    \n",
    "    def derivative(self, n=1):\n",
    "        assert n >= 0\n",
    "        if n > 0:\n",
    "            return self._derivative().derivative(n - 1)\n",
    "        return self\n",
    "    \n",
    "    def __add__( self, other ):\n",
    "        if isinstance( other, numbers.Real ):\n",
    "            other = Constant(other)\n",
    "        if isinstance( other, FunctionWithDerivative ):\n",
    "            if isinstance( other, Zero ):\n",
    "                return self\n",
    "            return Add( self, other )\n",
    "        return NotImplemented\n",
    "    \n",
    "    def __radd__( self, other ):\n",
    "        return self.__add__(other)\n",
    "    \n",
    "    def __mul__( self, other ):\n",
    "        if isinstance( other, numbers.Real ):\n",
    "            other = Constant(other)\n",
    "        if isinstance( other, FunctionWithDerivative ):\n",
    "            if isinstance( other, Zero ):\n",
    "                return Zero()\n",
    "            return Multiply( self, other )\n",
    "        return NotImplemented\n",
    "    \n",
    "    def __rmul__( self, other ):\n",
    "        return self.__mul__(other)\n",
    "    \n",
    "    def plot( self, interval=None ):\n",
    "        if interval is None:\n",
    "            interval = [0, 1]\n",
    "        x = np.linspace(*interval, 100)\n",
    "        y = [ self(x_) for x_ in x ]\n",
    "        from matplotlib import pyplot as plt\n",
    "        %matplotlib inline\n",
    "        plt.figure()\n",
    "        plt.plot( x, y )\n",
    "        plt.show();\n",
    "        \n",
    "    \n",
    "class Add( FunctionWithDerivative ):\n",
    "    \n",
    "    def __init__( self, func0, func1 ):\n",
    "        \n",
    "        assert \\\n",
    "            isinstance(func0, FunctionWithDerivative) \\\n",
    "            and isinstance(func1, FunctionWithDerivative)\n",
    "            \n",
    "        self.func0 = func0\n",
    "        self.func1 = func1\n",
    "        \n",
    "        function = lambda x: func0(x) + func1(x)\n",
    "        \n",
    "        super().__init__( function )\n",
    "        \n",
    "    def _derivative( self ):\n",
    "        return self.func0.derivative() + self.func1.derivative()\n",
    "    \n",
    "\n",
    "class Multiply( FunctionWithDerivative ):\n",
    "    \n",
    "    def __init__( self, func0, func1 ):\n",
    "        \n",
    "        # make sure both inputs are Functions with derivatives\n",
    "        assert \\\n",
    "            isinstance(func0, FunctionWithDerivative) \\\n",
    "            and isinstance(func1, FunctionWithDerivative)\n",
    "            \n",
    "        self.func0 = func0\n",
    "        self.func1 = func1\n",
    "        \n",
    "        function = lambda x: func0(x) * func1(x)\n",
    "        \n",
    "        super().__init__( function )\n",
    "        \n",
    "    def _derivative( self ):\n",
    "        return self.func0.derivative() * self.func1 + \\\n",
    "                                     self.func0 * self.func1.derivative()\n",
    "\n",
    "    \n",
    "class Constant( FunctionWithDerivative ):\n",
    "    \n",
    "    def __new__( cls, constant ):\n",
    "        \n",
    "        if constant == 0:\n",
    "            return Zero()\n",
    "        \n",
    "        return FunctionWithDerivative.__new__(cls)    \n",
    "    \n",
    "    def __init__( self, constant ):\n",
    "        \n",
    "        self.c = constant\n",
    "        function = lambda x: constant\n",
    "        super().__init__( function )\n",
    "\n",
    "    def _derivative( self ):\n",
    "        return Constant(0)\n",
    "\n",
    "    \n",
    "class Zero( FunctionWithDerivative ):\n",
    "    \n",
    "    def __init__( self ):\n",
    "        super().__init__( lambda x: 0 )\n",
    "        \n",
    "    def _derivative( self ):\n",
    "        return Zero()\n",
    "\n",
    "    \n",
    "\"\"\"\n",
    "    Exponentials ain't got no roots.\n",
    "\"\"\"\n",
    "        \n",
    "        \n",
    "class Exponential( FunctionWithDerivative ):\n",
    "    'exp(a*x)'\n",
    "    \n",
    "    def __new__( cls, a ):\n",
    "        if a == 0:\n",
    "            'a=0 -> exp(a*x) = 1'\n",
    "            return Constant(1)\n",
    "        return FunctionWithDerivative.__new__(cls)\n",
    "    \n",
    "    def __init__( self, a ):\n",
    "        self.a = a\n",
    "        function = lambda x: np.exp( self.a * x )\n",
    "        super().__init__( function )\n",
    "        \n",
    "    def _derivative( self ):\n",
    "        return self.a * Exponential(self.a)\n",
    "\n",
    "    \n",
    "\"\"\"\n",
    "    Polynomials can have roots.\n",
    "\"\"\"\n",
    "    \n",
    "    \n",
    "class Polynomial( FunctionWithDerivative, FindRootMixin ):\n",
    "    \n",
    "    def __new__( cls, prefactors ):\n",
    "        if len(prefactors) == 0:\n",
    "            return Zero()\n",
    "        if len(prefactors) == 1:\n",
    "            return Constant( prefactors[0] )\n",
    "        return FunctionWithDerivative.__new__(cls)\n",
    "    \n",
    "    def __init__( self, prefactors ):\n",
    "        self.prefactors = prefactors\n",
    "        function = lambda x: \\\n",
    "            sum( a * x ** i for i, a in enumerate( self.prefactors ) )\n",
    "        super().__init__( function )\n",
    "    \n",
    "    def _derivative( self ):\n",
    "        derivative_prefactors = [ a * i for i, a in enumerate(self.prefactors) ][1:]\n",
    "        return Polynomial( derivative_prefactors )\n",
    "    \n",
    "    \n",
    "\"\"\"\n",
    "    Sin / Cos functions have roots.\n",
    "\"\"\"\n",
    "\n",
    "    \n",
    "class Sin( FunctionWithDerivative, FindRootMixin ):\n",
    "    'sin(w*x)'\n",
    "    \n",
    "    def __new__( cls, w ):\n",
    "        if w == 0:\n",
    "            return Zero()\n",
    "        return FunctionWithDerivative.__new__(cls)\n",
    "    \n",
    "    def __init__( self, w ):\n",
    "        self.w = w\n",
    "        super().__init__( lambda x: np.sin( self.w * x) )\n",
    "        \n",
    "    def _derivative( self ):\n",
    "        return self.w * Cos( self.w )\n",
    "\n",
    "    \n",
    "class Cos( FunctionWithDerivative, FindRootMixin ):\n",
    "    'cos(w*x)'\n",
    "    \n",
    "    def __new__( cls, w ):\n",
    "        if w == 0:\n",
    "            return Constant(1)\n",
    "        return FunctionWithDerivative.__new__(cls)\n",
    "    \n",
    "    def __init__( self, w ):\n",
    "        self.w = w\n",
    "        super().__init__( lambda x: np.cos( self.w * x) )\n",
    "        \n",
    "    def _derivative( self ):\n",
    "        return -self.w * Sin( self.w )"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "prefactors = [ 1, -1, -2, 4 ]\n",
    "pol = Polynomial( prefactors )\n",
    "root = pol.find_root()\n",
    "\n",
    "\"Let's plot it\"\n",
    "%matplotlib inline\n",
    "from matplotlib import pyplot as plt\n",
    "x = np.linspace( -3, 3, 1000 )\n",
    "y = [ pol(x_) for x_ in x ]\n",
    "plt.plot( x, y )\n",
    "plt.scatter( [root], [0], c='r' )\n",
    "plt.show()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "###### Further reading: @property, pre-built class templates from the collections module, proper operator overloading\n",
    "###### Thank you for you attention !\n",
    "# Questions ?"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.7.1"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
